from __future__ import absolute_import
#
# denon.avr - Denon/Marantz AV Receiver control - main objects
#
# Copyright 2012-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import asyn
import asyn.utility
import asyn.scan
import math
from collections import deque
import time
import inspect

from . import core
from . import values
from . import attr
from .attr import Attribute, VolumeAttribute, BooleanAttribute, DictAttribute, EnumAttribute
from .attr import SEL_MAIN, SEL_ZONE


DEBUG=None
#def DEBUG(*it): print it


#
# Operational delays.
# The Denons drop commands and replies when they overrun their (small) buffers.
# We need to intelligently pace our requests to (largely) avoid that.
#
RDELAY = 0.020		# response pacing (per expected event)
IDELAY = 0.500		# initial command delay to help waking things up
CDELAY = 0.050		# default command-to-command delay


#
# All receiver features that are per-zone.
#
class Zone(attr.Device, asyn.Callable):
	""" A zone on a Denon AVR. The main zone is one of them. """
	zone_sel = SEL_ZONE

	def __init__(self, dev, id, callout=None):
		attr.Device.__init__(self)
		asyn.Callable.__init__(self, callout=callout)
		self.dev = dev				# host AVR device
		self.id = id				# zone id ('M', '2', etc.)

	#
	# Per-zone attributes. They apply to the main zone as well.
	#
	power =			BooleanAttribute(cmd=('ZM', ''), scan=(r'ZM(ON|OFF)', r'Z(\d)(ON|OFF)'), pri=10)
	source =		Attribute(cmd=('SI', ''), scan=(r'SI([^\r]+)', r'Z(\d)([^\r]+)'), pri=5)
	volume =		VolumeAttribute(cmd=('MV', ''), scan=(r'MV(\d+)', r'Z(\d)(\d+)'), zero=core.VOL_ZERO, fine=True, pri=8)
	muted =			BooleanAttribute(cmd=('MU', 'MU'), scan=(r'MU(ON|OFF)', r'Z(\d)MU(ON|OFF)'), pri=8)
	base =			VolumeAttribute(cmd=('PSBAS ', 'PSBAS '), scan=(r'PSBAS (\d+)', r'Z(\d)PSBAS (\d+)'))
	treble =		VolumeAttribute(cmd=('PSTRE ', 'PSTRE '), scan=(r'PSTRE (\d+)', r'Z(\d)PSTRE (\d+)'))
	channel_volume = DictAttribute(scan=(r'CV(\w+) (\d+)', r'Z(\d)CV(\w+) (\d+)'),
						query=('CV?', 'CV?'), convert=values.DenonVolume(zero=core.CHAN_ZERO), pri=-5, count=8)
	quick_select =	Attribute(type=int, cmd=('MSQUICK', 'QUICK'), query=('MSQUICK ?', 'QUICK ?'),
						scan=(r'MSQUICK(\d+)', r'Z(\d)QUICK(\d+)'))


	#
	# Explicitly (re)query an attribute
	#
	def query(self, name):
		if name not in type(self).attributes:
			raise AttributeError('no "%s" attribute in AVR' % name)
		attr = type(self).attributes[name]
		if attr.query is None:
			raise AttributeError('"%s" cannot be queried' % name)
		self._query(attr)

	def _query(self, attr):
		query = attr.query[self.zone_sel]
		for filter in attr.edit_send:
			query = filter(self, query)
		self.command(query, delay=attr.count*RDELAY)


	def _update(self, attr, args):
		""" Record an AVR-reported change to a zone attribute. """
		old_value = getattr(self, attr.name)
		attr.update(self, *args)
		new_value = getattr(self, attr.name)
		if attr.raw or new_value != old_value:
			self.callout('change', attr, new_value, old_value)

	def command(self, cmd, **kwargs):
		""" Send a zone command (ZnWhatever) for this zone. """
		self.dev.command("Z%s%s" % (self.id, cmd), **kwargs)


	#
	# Explicit commands
	#
	_quick_save = ('MSQUICK%d MEMORY', 'QUICK%d MEMORY')
	def save_quick_select(self, slot):
		self.command(self._quick_save[self.zone_sel] % slot)


#
# Here is the object for a whole AVR box
#
class AVR(Zone, asyn.utility.Idler):
	""" An actual Denon AVR device.

		This IS a Zone (the "main zone", id='M'). For AVRs that have no other zones,
		this is the only one. Note that some Denon AVRs respond (harmlessly) to zone
		commands for zones that it doesn't electrically support.

		The connection to the AVR can be either via TCP or serial link. The protocols
		are identical. Note that a network-connected Denon will serve (only) one network
		and one serial stream simultaneously.

		Denon has no identification or introspection support. We start out knowing
		nothing about the device. Changes are reported fairly reliably. Either call
		discover() to start the device discovery procedure, or selectively ask for
		just what you need.
	"""
	zone_sel = SEL_MAIN

	def __init__(self, control, callout=None):
		""" Construct an AVR given an asyn.Controller. You need to open() it to use it. """
		Zone.__init__(self, self, 'M', callout=callout)
		asyn.utility.Idler.__init__(self, control)
		self._io = None
		self.control = control
		self.zones = { 'M': self }		# the main zone is always there
		self.radio_type = None			# HD (HDRADIO) or AN (analog)
		self._cmd_queue = deque()		# outgoing command queue
		self._cmd_timer = None			# timer for command pacing
		self.discovered = False			# no discovery completed yet
		self.discovering = False		# auto-discovery active
		self._discover_remnants = None	# what auto-discovery couldn't discover
		self._zone_discovery = 'new'	# state of zone discovery (new/seeded/done)


	#
	# Screen hybrid tuner/hdradio events, establishing the radio type.
	#
	def radio_cmd(self, cmd):
		if self.radio_type:
			cmd = cmd.replace("HD", self.radio_type, 1)
		return cmd

	def radio_filter(self, args):
		""" Process radio events that start with an (HD|AN) element. """
		radio = args[0]
		if radio == "AN" and self.radio_type == "HD":	# do not downgrade; eat this
			return None
		self.radio_type = radio			# record type
		return args[1:]


	#
	# Attributes for a whole AVR (not a zone). See attr.py.
	#
	main_power = 		BooleanAttribute(cmd='PW', off='STANDBY', raw=True)

	# switching
	video_source =		Attribute(cmd='SV', pri=2)
	record_source =		Attribute(cmd='SR', pri=2)
	hdmi_monitor =		Attribute(cmd='VSMONI', query='VSMONI ?')
	audio_output =		EnumAttribute(cmd='VSAUDIO ', choices=('AMP', 'TV'))

	# sound processing
	surround =			Attribute(cmd='MS')
	surround_mode =		Attribute(cmd='PSMODE:', scan='PSMODE:([^\r]+)', query='PSMODE: ?')
	front_speakers =	EnumAttribute(cmd='PSFRONT ', choices=('SPA', 'SPB', 'SPA+B'), query='PSFRONT?', pri=1)
	reference_level =	EnumAttribute(type=int, cmd='PSREFLEV ', choices=(0, 5, 10, 15))
	dynamic_volume =	EnumAttribute(cmd='PSDYNVOL ', choices=('DAY', 'EVE', 'NGT'))
	max_volume =		VolumeAttribute(scan=r'MVMAX (\d+)', zero=core.VOL_ZERO)
	mult_eq =			Attribute(cmd='PSMULTEQ:', query='PSMULTEQ: ?')
	dynamic_eq =		BooleanAttribute(cmd='PSDYNEQ ')
	cinema_eq =			BooleanAttribute(cmd='PSCINEMA EQ.', query='PSCINEMA EQ. ?')
	tone_control =		BooleanAttribute(cmd='PSTONE CTRL ')
	dcomp =				EnumAttribute(cmd='PSDCO ', choices=('OFF', 'LOW', 'MID', 'HIGH'))
	drc =				EnumAttribute(cmd='PSDRC ', choices=('OFF', 'LOW', 'MID', 'HIGH', 'AUTO'))
	audio_restorer =	EnumAttribute(cmd='PSRSTR ', choices=('OFF', 'MODE1', 'MODE2', 'MODE3'))

	# video
	digital_input_type = EnumAttribute(cmd='SD', choices=('AUTO', 'HDMI', 'DIGITAL', 'ANALOG'))
	digital_input_mode = EnumAttribute(cmd='DC', choices=('AUTO', 'PCM', 'DTS'))
	resolution =		Attribute(cmd='VSSC', query='VSSC ?')
	hdmi_resolution =	Attribute(cmd='VSSCH', query='VSSCH ?')
	video_aspect =		EnumAttribute(cmd='VSASP', query='VSASP ?', choices=('NRM', 'FUL'))
	video_processing_mode = EnumAttribute(cmd='VSVPM', query='VSVPM ?', choices=('AUTO', 'GAME', 'MOVI'))

	# tuner-related
	tuner_frequency =	Attribute(type=float, cmd='TFHD', scan='TF(HD|AN)(\d{6})', convert=values.Frequency(), edit_send=[radio_cmd], edit_update=[radio_filter])
	tuner_band =		EnumAttribute(cmd='TMHD', scan='TM(HD|AN)(AM|FM)', choices=('AM', 'FM'), edit_send=[radio_cmd], edit_update=[radio_filter], pri=2)
	tuner_hd_mode =		EnumAttribute(cmd='TMHD', choices=('AUTOHD', 'AUTO', 'MANUAL', 'ANAAUTO', 'ANAMANU'))
	tuner_preset =		Attribute(cmd='TPHD', scan='TP(HD|AN)(\w+)', edit_send=[radio_cmd], edit_update=[radio_filter])
	tuner_mc_channel =	Attribute(type=int, cmd='TFHDMC', scan='TFHDMC(\d+)', query='TFHD?')
	tuner_hd =			DictAttribute(query='HD?', pri=1,
		scan='HD(ST NAME|STL NAME|SIG LEV|MLT CURRCH|MLT CAST CH|\w+) ([^\r]+)')

	# miscellaneous and WTF
	# front_preset can be set and announces when engaged, but you can't query it
	front_preset =		Attribute(type=int, cmd='NSP', default=0)
	# network_preset is write-only
	network_preset =	Attribute(cmd='NSB', scan=False, query=False, convert=values.InternetPreset())

	# read-only lists
	network_program_list = DictAttribute(scan='NSH(\d{2})([^\r]+)', query='NSH',
		convert=values.StringConverter(), key_convert=values.InternetPreset(), count=56, pri=None)
	front_preset_list =	DictAttribute(key_type=int, scan='NSP(\d{2})([^\r]+)', query='NSP', convert=values.StringConverter(), count=3, pri=None)
	nse =				DictAttribute(key_type=int, scan='NSE(\d)([^\r]+)', query='NSE',
		convert=values.StringConverter(), count=8, pri=None)


	#
	# Open the device.
	#
	def open(self, file):
		""" Establish contact with the device.

			The file argument must be based on a file descriptor - it can be either
			a genuine File object, or an integer file descriptor. Both sockets and
			local streams are acceptable.

			This method does not start discovery. Call discover() after it if you want it.
		"""
		self._io = self.control.stream(file, callout=self._event)
		self._io.scan = self._make_scan()
		self.open_time = time.time()
		self.xmit_idle()


	#
	# Close down the object
	#
	def close(self):
		""" Release system resources. """
		self.idle_cancel()
		if self._io:
			self._io.close()
			self._io = None
		if self._cmd_timer:
			self._cmd_timer.cancel()


	#
	# Trigger operations.
	# Triggers can be condition-programmed in the receiver;
	# but once you send them an explicit on/off command, they "stick"
	# until the receiver powers down. Slightly weird but useful.
	# There is no way to query the state of a trigger.
	#
	def set_trigger(self, trigger_number, state):
		self.command("TR%d %s" % (trigger_number, "ON" if state else "OFF"))


	#
	# Idle trigger
	#
	def idle(self):
		""" Issue a no-op command to check connectivity. """
		self.command("MV?")

	def idle_timeout(self):
		self.callout("timeout")


	#
	# Find and list zones. After zone discovery is finished, all recognized zones
	# are found in the zones dict.
	#
	def _zone(self, id):
		""" Return a named zone. Create it if necessary. Use 'M' for the main zone. """
		if id in self.zones:
			zone = self.zones[id]
		else:
			zone = self.zones[id] = Zone(self, id)
		return zone

	def zone_list(self, primary=False):
		""" A list of known zones, in reasonable order (M, 2, ...). """
		if primary:
			yield self
		for z in core.ZONE_CANDIDATES:
			if z in self.zones:
				yield self.zones[z]


	#
	# Construct the Selectable scanner for events from the AVR.
	# We pick up scan rules from attribute selectors, and add some
	# non-attribute rules at the end.
	#
	scans = [	# events not mapped to an Attribute
		(r'TR(\d+) (\w+)\r', 'trigger_response'),
		(r'(.*?)\r', 'unknown'),
	]

	def _make_scan(self):
		scans = [(attr.scan[sel] + r'\r', attr, sel)
			for attr in self.attributes.values()
			for sel in [SEL_MAIN, SEL_ZONE]
			if attr.scan and attr.scan[sel]] + self.scans
		# sort by regex, putting ([\ after letters (wildcards last)
		xl = list(range(0, 256))
		xl[ord('\\')] = 254
		xl[ord('(')] = xl[ord('[')] = 255
		collate = ''.join(map(chr, xl))
		scans.sort(key=lambda a: a[0].translate(collate))
		return asyn.scan.Regex(scans)


	#
	# Master event dispatch.
	#
	def _event(self, ctx, *args):
		if ctx.error:
			return self.callout(ctx)
		if ctx.state == 'END':
			return self.callout(ctx)
		elif ctx.state == 'CLOSE':
			return
		assert ctx.scan
		if DEBUG: DEBUG('<-', ctx, *args)
		if isinstance(ctx.state, Attribute):
			self.callout("receive", ctx.state.name, *args)
			for filter in ctx.state.edit_update:
				args = filter(self, args)
				if args is None:	# filter said, "abort!"
					return
			if ctx.aux[0] == SEL_ZONE:
				self._zone(args[0])._update(ctx.state, args[1:])
			else:
				self._update(ctx.state, args)
		elif ctx.state == 'trigger_response':
			(tnum, s) = args
			self.callout("trigger", int(tnum), s == "ON")
		else:
			self.callout("receive", ctx.state, *args)
		self.idle_activity()


	#
	# Discovery management.
	# This is painful. The Denons report no errors of any kind, and drop commands
	# and replies whenever their puny little buffers overflow. They also tend to
	# go nap whenever they feel like it. Their interface specification claims 200ms
	# response; but if we waited 200ms after each command, we'd take minutes to
	# properly discover all state.
	# So we compromise by shooting out queries with some pacing heuristic, keep track
	# of what attributes we haven't received yet, and retry the missing ones until
	# there seems no more progress. This works decently in practice, and gets the job
	# done in a few seconds.
	# Discovery as done here does not change the state of the receiver at all.
	#
	def discover(self):
		""" Request passive discovery of yet-unknown attributes in all known zones. """
		self.command("TFHD?", delay=IDELAY)	# first query (extra slow; wakey wakey)
		self.command("TFAN?")	# in case it has an analog radio
		self._schedule_discovery()

	# construct and schedule a discovery run
	def _schedule_discovery(self):
		self._zone_discover()
		work = [(zone, attr)
			for zone in self.zone_list(primary=True)
			for attr in zone.missing_attributes()
			if attr.pri is not None and attr.query and attr.query[self.zone_sel]]
		work.sort(key=lambda s: s[1].pri, reverse=True)
		if work == self._discover_remnants:		# stalled, declare victory or defeat
			self.discovering = False
			self._discover_remnants = None
			if self.zones:
				self.discovered = True
				self.callout('discover', 'done')
			else:
				self.callout('discover', 'failed')
		else:	# still making progress, schedule another batch
			self._discover_tasks = work
			self._discover_remnants = list(work)
			self.callout('discover', 'restart' if self.discovering else 'start')
			self.discovering = True
			self._cmd_push()

	def _next_discovery(self):
		""" Locate the next queued query command and issue it. """
		try:
			while 1:
				(zone, attr) = self._discover_tasks.pop(0)
				if attr.name in zone._attr_values:
					continue	# already resolved
				self.callout('discover', 'attr', zone, attr)
				zone._query(attr)
				return
		except IndexError:
			self._schedule_discovery()

	# Zone discovery is interleaved with attribute discovery.
	# We discover zones by receiving events about them. At the start, we ask about
	# all possible zones (M, 2-9). Any zone that reports a volume is considered present.
	def _zone_discover(self):
		""" Zone discovery management. Interleaved with attribute discovery. """
		if self._zone_discovery == 'new':
			for id in core.ZONE_CANDIDATES:
				self.command("Z%s?" % id, delay=6*RDELAY)
			self._zone_discovery = 'seeded'
		elif self._zone_discovery == 'seeded':
			for zone in list(self.zones.keys()):
				if not self.zones[zone].has_attribute('volume'):
					del self.zones[zone]
			self.callout('discover', 'zones')
			self._zone_discovery = 'done'

	# called when a command could be sent but the transmit queue is empty
	def xmit_idle(self):
		if self.discovering:
			self._next_discovery()


	#
	# Core command management
	#
	def command(self, cmd, delay=CDELAY):
		""" Issue a (raw text) command to the AVR.

			Outgoing commands are paced so as not to overload the receiver.
			This is entirely asynchronous; we do not wait for replies.
		"""
		self._cmd_queue.append((cmd, delay))	# queue it
		self._cmd_push()

	def _cmd_push(self):
		if not self._cmd_timer:			# if idle...
			self._cmdxmit()				# ... send it

	def _cmdxmit(self, ctx=None):
		try:
			(cmd, delay) = self._cmd_queue.popleft()
			self.callout('send', cmd)
			if DEBUG: DEBUG('->', cmd)
			self._io.write_a(cmd + '\r')
			self._cmd_timer = self.control.schedule(self._cmdxmit, after=delay)
		except IndexError:
			self._cmd_timer = None
			self.xmit_idle()
